#!/usr/bin/env python

import re
import json
import fnmatch
import requests
import urllib
from pathlib import Path
from .exceptions import (
    SpaceNotFound,
    FolderNotFound,
    ListNotFound,
    TaskNotFound,
    ClickUpException,
    GenericException,
)

BRANCH_SEPARATOR = "-"
SERVER_URL = "https://api.clickup.com/api/v2/"

__all__ = ["ClickUpClient"]


class ClickUpClient:
    def __init__(self, config):
        self.server = SERVER_URL
        self.config = config
        self.team_id = self.config.default_clickup_team_id
        self.workspace = None

    def send_request(self, part, method="GET", request_args=None, payload=None, **kwargs):
        "Send HTTP Request to ClickUP"
        part = part.format(**kwargs)
        url = urllib.parse.urljoin(self.server, part)
        headers = {"Authorization": self.config.default_clickup_token}
        request_args = request_args or dict()
        if payload is not None:
            request_args["json"] = payload
        response = requests.request(method=method, url=url, headers=headers, **request_args)
        # self.save_response(response)
        payload = response.json()
        if "err" in payload:
            if self.config.is_verbose():
                print(json.dumps(payload, indent=2))
            raise ClickUpException(payload["err"])
        return payload

    def save_response(self, response):
        rqs = response.request
        path_url = rqs.path_url.strip("/").replace("..", "").split("?")[0] + "." + rqs.method.lower()
        filename = Path.cwd() / "tests" / "data" / Path(*path_url.split("/"))
        print(filename)
        filename.parent.mkdir(parents=True, exist_ok=True)
        with filename.open("wb") as f:
            f.write(response.content)
        headers_filename = filename.parent / (filename.name + ".headers")
        with headers_filename.open("w") as f:
            headers = dict(response.headers)
            headers["__Status__"] = response.status_code
            json.dump(headers, f, indent=2)

    def get_user(self):
        return self.send_request("user")["user"]

    def get_teams(self):
        return self.send_request("team")["teams"]

    def get_space_by_id(self, space_id):
        "Get a space by id"
        try:
            data = self.send_request(f"space/{space_id}/")
            return Space(self, data)
        except ClickUpException:
            raise SpaceNotFound(f"Space '{space_id}' not found")

    def get_list_by_id(self, list_id):
        "Get a list by id"
        try:
            data = self.send_request(f"list/{list_id}/")
            return List(self, data)
        except ClickUpException:
            raise ListNotFound(f"List '{list_id}' not found")

    def get_task_by_id(self, task_id):
        "Get a task by id"
        try:
            task_id = task_id.lstrip("#")
            data = self.send_request(f"task/{task_id}/")
            return Task(self, data)
        except ClickUpException:
            raise TaskNotFound(f"Task '{task_id}' not found")

    def get_task(self, task_id, lst=None, space=None, folder=None):
        "Get a task by id"
        if not task_id:
            return None
        else:
            return self.get_task_by_id(task_id)

    def get_workspace(self, index=0):
        "Get the workspace"
        if self.workspace is None:
            teams = self.get_teams()
            data = teams[index]
            self.workspace = Workspace(self, data)
        return self.workspace

    def get_space(self, space_id_or_name):
        "Get a space by name or id"
        if not space_id_or_name:
            return None
        elif space_id_or_name.isdigit():  # get by id
            return self.get_space_by_id(space_id=space_id_or_name)
        else:  # get by name
            spaces = self.get_workspace().get_spaces()
            try:
                return [space for space in spaces if space.get("name") == space_id_or_name][0]
            except Exception:
                raise SpaceNotFound(f"Space '{space_id_or_name}' not found")

    def get_folder(self, folder_id_or_name, space=None):
        "Get a folder by name or id"
        if not folder_id_or_name:
            return None
        elif folder_id_or_name.isdigit():  # get by id
            return self.get_folder_by_id(folder_id=folder_id_or_name)
        elif space is None:
            raise GenericException("Please specify space")
        else:  # get by name
            return space.get_folder_by_name(name=folder_id_or_name)

    def get_folder_by_id(self, folder_id):
        "Get a folder by id"
        try:
            data = self.send_request(f"folder/{folder_id}/")
            return Folder(self, data)
        except ClickUpException:
            raise FolderNotFound(f"Folder '{folder_id}' not found")

    def get_list(self, list_id_or_name, space=None, folder=None):
        "Get a list by name or id"
        if not list_id_or_name:
            return None
        elif list_id_or_name.isdigit():
            return self.get_list_by_id(list_id=list_id_or_name)
        else:
            lists = []
            if not space and not folder:
                raise GenericException("Please specify space/folder")
            if space:
                lists.extend(space.get_space_lists())
            if folder:
                lists.extend(folder.get_folder_lists())
            try:
                return [lst for lst in lists if lst.get("name") == list_id_or_name][0]
            except Exception:
                raise ListNotFound(f"List '{list_id_or_name}' not found")

    def query(
        self,
        space=None,
        folder=None,
        lst=None,
        task=None,
        filter_type=None,
        filter_name=None,
        hierarchy=False,
    ):
        "Get spaces/folders/lists/tasks"
        result = []
        # Workspace
        if hierarchy:
            workspace = self.get_workspace()
            result.append(workspace)
        # Get space
        space = self.get_space(space)
        if space and hierarchy:
            result.append(space)
        # Get folder
        folder = self.get_folder(folder, space=space)
        if folder and hierarchy:
            if space is None:
                space = Space(self, folder["space"])
                result.append(space)
            result.append(folder)
        # Get list
        lst = self.get_list(lst, space=space, folder=folder)
        if lst and hierarchy:
            if space is None:
                space = Space(self, lst["space"])
                result.append(space)
            if folder is None and lst.get("folder") and not lst["folder"].get("hidden"):
                folder = Folder(self, lst["folder"])
                result.append(folder)
            result.append(lst)
        # Get task
        task = self.get_task(task, lst=lst, space=space, folder=folder)
        if task and hierarchy:
            if space is None:
                space = self.get_space_by_id(task["space"]["id"])
                result.append(space)
            if folder is None and task.get("folder") and not task["folder"].get("hidden"):
                folder = Folder(self, task["folder"])
                result.append(folder)
            if lst is None and task.get("list"):
                lst = List(self, task["list"])
                result.append(lst)
            result.append(task)
        # Prepare result
        if task:
            # Task subtasks
            result.extend(task.get_subtasks())
        elif lst:
            # List tasks
            result.extend(lst.get_list_tasks())
        elif folder:
            # Folders lists
            result.extend(folder.get_folder_lists())
        elif space:
            # Space folders
            result.extend(space.get_space_folders())
            # Space folderless lists
            result.extend(space.get_space_lists())
        else:
            # Spaces
            result.extend(self.get_workspace().get_spaces())
        # Filter result by type
        if filter_type:
            result = [x for x in result if x["type"] == filter_type]
        # Filter result by name
        if filter_name:
            match = lambda x: fnmatch.fnmatch(x["name"].lower(), filter_name.lower())
            result = [x for x in result if match(x)]
        return result


class Workspace(dict):
    def __init__(self, client, data):
        self.update(data)
        self.team_id = data["id"]
        self.client = client
        self["type"] = "Workspace"
        self["label"] = self["type"]

    def get_spaces(self, archived=False):
        "Get spaces"
        result = self.client.send_request(f"team/{self.team_id}/space?archived={archived}")
        return [Space(self.client, space) for space in result["spaces"]]

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError(f"No such attribute: {name}")

    def __eq__(self, other) -> bool:
        return other is not None and isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)


class Space(dict):
    def __init__(self, client, data):
        self.update(data)
        self.space_id = data["id"]
        self.client = client
        self["type"] = "Space"
        self["label"] = self["type"]

    def get_folder_by_name(self, name):
        "Get space folder by name"
        folders = self.get_space_folders()
        try:
            return [folder for folder in folders if folder.get("name") == name][0]
        except Exception:
            raise FolderNotFound(f"Folder '{name}' not found")

    def get_space_folders(self, archived=False):
        "Get space folders"
        response = self.client.send_request(
            f"space/{self.id}/folder?archived={archived}",
        )
        return [Folder(self.client, data) for data in response["folders"]]

    def get_space_lists(self, archived=False):
        "Get space lists"
        response = self.client.send_request(
            f"space/{self.id}/list?archived={archived}",
        )
        return [List(self.client, data) for data in response["lists"]]

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError(f"No such attribute: {name}")

    def __eq__(self, other) -> bool:
        return other is not None and isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)


class Folder(dict):
    def __init__(self, client, data):
        self.update(data)
        self.client = client
        self["type"] = "Folder"
        self["label"] = self["type"]

    def get_folder_lists(self, archived=False):
        "Get folder lists"
        response = self.client.send_request(
            f"folder/{self.id}/list?archived={archived}",
        )
        return [List(self.client, data) for data in response["lists"]]

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError(f"No such attribute: {name}")

    def __eq__(self, other) -> bool:
        return other is not None and isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)


class List(dict):
    def __init__(self, client, data):
        self.update(data)
        self.client = client
        self["type"] = "List"
        self["label"] = self["type"]

    def get_list_tasks(self, include_closed=False):
        response = self.client.send_request(f"list/{self.id}/task?include_closed={include_closed}")
        return [Task(self.client, data) for data in response["tasks"]]

    def get_statuses(self):
        return [x["status"] for x in self.get("statuses")]

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError(f"No such attribute: {name}")

    def __eq__(self, other) -> bool:
        return other is not None and isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)


class Task(dict):
    def __init__(self, client, data):
        self.update(data)
        self.task_id = data["id"]
        self.client = client
        self["type"] = "Subtask" if self.get("parent") else "Task"
        self["label"] = self.status

    def update_task(self, **kargs):
        "Update task"
        self.client.send_request(
            f"task/{self.id}/",
            method="PUT",
            payload=kargs,
        )
        self.update(kargs)

    def post_task_comment(self, comment_text, notify_all=None, assignee=None):
        "Create task command"
        payload = {"comment_text": comment_text}
        if notify_all is not None:
            payload["notify_all"] = notify_all
        if assignee is not None:
            payload["assignee"] = assignee
        return self.client.send_request(
            f"task/{self.id}/comment",
            method="POST",
            payload=payload,
        )

    def get_subtasks(self, include_closed=False):
        "Get subtasks"
        response = self.client.send_request(f"list/{self.list['id']}/task?subtasks=true&include_closed={include_closed}")
        return [Task(self.client, data) for data in response["tasks"] if data.get("parent") == self.id]

    def has_subtasks(self, include_closed=False):
        "Returns true if the task has subtasks"
        return any(self.get_subtasks(include_closed))

    @property
    def branch_name(self):
        "Branch name"
        name = self.get("name")
        branch_name = BRANCH_SEPARATOR.join(
            (
                self.task_id.strip("#"),
                re.sub(r"[^A-Za-z0-9_\-]+", "", name.replace(" ", BRANCH_SEPARATOR)).lower(),
            )
        )
        return branch_name

    @property
    def status(self):
        "Task status"
        return self.get("status", {}).get("status", "-")

    def get_space(self):
        "Get task space"
        space_id = self.get("space", {}).get("id")
        return self.client.get_space_by_id(space_id)

    def get_list(self):
        "Get task list"
        list_id = self.get("list", {}).get("id")
        return self.client.get_list_by_id(list_id)

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError(f"No such attribute: {name}")

    def __eq__(self, other) -> bool:
        return other is not None and isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)



#!/usr/bin/env python

from pathlib import Path

try:
    from functools import cached_property
except ImportError:
    from backports.cached_property import cached_property
try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources
from .clickup import ClickUpClient
from .config import Config
from .git import Git
from .github import GitHubClient


__all__ = ["Workflow", "VERSION"]

VERSION = pkg_resources.read_text("alkemy_workflow", "VERSION")


class Workflow:
    def __init__(self, base_path=None, credentials_path=None):
        self.base_path = Path(base_path) if base_path else None
        self.credentials_path = Path(credentials_path) if credentials_path else None

    @cached_property
    def config(self):
        return Config(base_path=self.base_path, credentials_path=self.credentials_path)

    @cached_property
    def git(self):
        return Git(self.config)

    @cached_property
    def github(self):
        return GitHubClient(self.config)

    @cached_property
    def client(self):
        return ClickUpClient(self.config)



#!/usr/bin/env python

import subprocess
from pathlib import Path
from .exceptions import GitException


__all__ = ["Git"]


class Git:
    def __init__(self, config):
        self.config = config

    def run(self, *args, git_dir=None):
        git_dir = git_dir or (self.config and self.config.git_dir) or None
        if git_dir:
            args = ["git", "-C", git_dir] + list(args)
        else:
            args = ["git"] + list(args)
        completed_process = subprocess.run(args, capture_output=True)
        if completed_process.returncode != 0:
            stdout = completed_process.stdout.decode("utf-8")
            stderr = completed_process.stderr.decode("utf-8")
            raise GitException(f"git error: {stdout}\n{stderr}")
        return completed_process.stdout.strip().decode("utf-8")

    def run_gh(self, *args, git_dir=None):
        git_dir = git_dir or (self.config and self.config.git_dir) or None
        args = ["gh"] + list(args)
        completed_process = subprocess.run(args, capture_output=True, cwd=git_dir)
        if completed_process.returncode != 0:
            stdout = completed_process.stdout.decode("utf-8")
            stderr = completed_process.stderr.decode("utf-8")
            raise GitException(f"gh error: {stdout}\n{stderr}")
        return completed_process.stdout.strip().decode("utf-8")

    def get_current_branch(self):
        "Get current branch"
        return self.run("rev-parse", "--abbrev-ref", "HEAD")

    def get_toplevel(self, git_dir=None):
        "Get the path of the top-level directory of the working tree"
        return Path(self.run("rev-parse", "--show-toplevel", git_dir=git_dir))

    def get_remote_url(self):
        "Get remote url"
        try:
            return self.run("remote", "get-url", "origin")
        except GitException:
            return None

    def checkout(self, *args):
        "Switch branch"
        return self.run("checkout", *args)

    def pull(self):
        "Fetch from remote(if any)"
        if self.get_remote_url():
            return self.run("pull")

    def push(self, *args):
        "Update remote"
        return self.run("push", *args)

    def create_branch(self, branch_name, base_branch=None):
        "Create a new branch a switch to it"
        base_branch = base_branch or self.config.git_base_branch
        branch_already_exists = False
        self.checkout(base_branch)
        self.pull()
        try:
            self.checkout("-b", branch_name)
        except GitException:
            branch_already_exists = True
            self.checkout(branch_name)
        try:
            self.push("--set-upstream", "origin", branch_name)
        except GitException:
            pass
        return branch_already_exists

    def get_github_url(self, branch_name, remote_url=None):
        "Get link to github (if origin is github)"
        url = remote_url or self.get_remote_url()
        if not url or "github.com" not in url:
            return None
        if url.startswith("git@github.com:"):
            url = url.replace("git@github.com:", "https://github.com/")
        if url.endswith(".git"):
            url = url[:-4]
        if branch_name:
            url = url + "/tree/" + branch_name
        return url



#!/usr/bin/env python

import json
import urllib
import requests
from pathlib import Path
from .exceptions import GitHubException

REPO_BASE_URL = "https://github.com/"
SERVER_URL = "https://api.github.com/"

__all__ = ["GitHubClient"]


class GitHubClient:
    def __init__(self, config):
        self.server = SERVER_URL
        self.config = config

    def send_request(self, part, method="GET", request_args=None, payload=None, **kwargs):
        "Send HTTP Request to GitHub"
        part = part.format(**kwargs)
        url = urllib.parse.urljoin(self.server, part)
        headers = {
            "Accept": "application/vnd.github+json",
            "Authorization": f"token {self.config.default_github_token}",
        }
        request_args = request_args or dict()
        if payload is not None:
            request_args["json"] = payload
        response = requests.request(method=method, url=url, headers=headers, **request_args)
        # self.save_response(response)
        payload = response.json()
        if self.config.is_verbose():
            print(json.dumps(payload, indent=2))
        if response.status_code < 200 or response.status_code > 299:
            errors = "\n".join([error["message"] for error in payload.get("errors", []) if error.get("message")])
            if errors:
                raise GitHubException(f"GitHub error: {payload['message']}\n{errors}")
            else:
                raise GitHubException(f"GitHub error: {payload['message']}")
        return payload

    def save_response(self, response):
        rqs = response.request
        path_url = rqs.path_url.strip("/").replace("..", "").split("?")[0] + "." + rqs.method.lower()
        filename = Path.cwd() / "tests" / "data" / Path(*path_url.split("/"))
        print(filename)
        filename.parent.mkdir(parents=True, exist_ok=True)
        with filename.open("wb") as f:
            f.write(response.content)
        headers_filename = filename.parent / (filename.name + ".headers")
        with headers_filename.open("w") as f:
            headers = dict(response.headers)
            headers["__Status__"] = response.status_code
            json.dump(headers, f, indent=2)

    def extract_repo(self, repo_url):
        if not repo_url.startswith(REPO_BASE_URL):
            raise GitHubException(f"Invalid repository URL, must starts with {REPO_BASE_URL}")
        repo_url = repo_url[len(REPO_BASE_URL) :]
        if repo_url.endswith(".git"):
            repo_url = repo_url[:-4]
        return repo_url

    def get_ref(self, repo_url, ref):
        "Get a reference"
        repo = self.extract_repo(repo_url)
        return self.send_request(f"repos/{repo}/git/refs/heads/{ref}")

    def create_ref(self, repo_url, branch_name, sha):
        "Create a reference"
        repo = self.extract_repo(repo_url)
        return self.send_request(
            f"repos/{repo}/git/refs",
            method="POST",
            payload={"ref": f"refs/heads/{branch_name}", "sha": sha},
        )

    def create_branch(self, repo_url, branch_name, base_branch=None):
        "Create a new remote branch"
        base_branch = base_branch or self.config.git_base_branch
        # Check if branch already exists
        try:
            ref_response = self.get_ref(repo_url, branch_name)
            return True  # branch already exists
        except GitHubException:
            pass
        # Get base branch ref
        ref_response = self.get_ref(repo_url, base_branch)
        # Create the new branch
        self.create_ref(repo_url, branch_name, ref_response["object"]["sha"])
        return False  # new branch

    def create_pull_request(self, repo_url, branch_name, title, base_branch=None):
        "Create a new pull request"
        base_branch = base_branch or self.config.git_base_branch
        repo = self.extract_repo(repo_url)
        payload = {"title": title, "head": branch_name, "base": base_branch}
        return self.send_request(f"repos/{repo}/pulls", method="POST", payload=payload)

    def list_pull_request(self, repo_url):
        "List pull request"
        repo = self.extract_repo(repo_url)
        return self.send_request(f"repos/{repo}/pulls")

    def merge_pull_request(self, repo_url, pr_nr, base_branch=None):
        "Create a new pull request"
        base_branch = base_branch or self.config.git_base_branch
        repo = self.extract_repo(repo_url)
        return self.send_request(f"repos/{repo}/pulls/{pr_nr}/merge", method="PUT")

    def get_user(self):
        return self.send_request("user")

    def list_issues(self, repo_url):
        "List repository issues"
        repo = self.extract_repo(repo_url)
        return self.send_request(f"repos/{repo}/issues", per_page=100)



#!/usr/bin/env python

import os
import configparser
import tempfile
from pathlib import Path
from .exceptions import ConfigException, GitException
from .git import Git


__all__ = ["Config"]

AW_VERBOSE = "AW_VERBOSE"
CLICKUP_TOKEN = "CLICKUP_TOKEN"
GITHUB_TOKEN = "GITHUB_TOKEN"
DEFAULT_GIT_BASE_BRANCH = "main"
CLICKUP_STATUS_IN_PROGRESS = "in_progress"
CLICKUP_STATUS_PR = "in_review"
CLICKUP_STATUS_MA = "done"
CREDENTIALS_KEYS = (
    "default_clickup_token",
    "default_clickup_team_id",
    "default_github_token",
)
CONFIG_KEYS = (
    "git_base_branch",
    "clickup_status_in_progress",
    "clickup_status_pr",
    "clickup_status_ma",
)


class Config:

    config_path = None
    default_clickup_token = None
    default_clickup_team_id = None
    default_github_token = None
    git_dir = None
    git_base_branch = DEFAULT_GIT_BASE_BRANCH
    clickup_status_in_progress = CLICKUP_STATUS_IN_PROGRESS
    clickup_status_pr = CLICKUP_STATUS_PR
    clickup_status_ma = CLICKUP_STATUS_MA

    def __init__(self, base_path=None, credentials_path=None):
        # Load credentials
        self.load_credentials(credentials_path)
        # Get project config
        try:
            self.load_config(base_path)
        except GitException:
            self.git_dir = None
            self.config_path = None

    def load_credentials(self, credentials_path=None):
        "Load credentials - load token from CLICKUP_TOKEN environment variable or ~/.alkemy_workflow/credentials file"
        credentials_path = credentials_path or self.get_credentials_path()
        if credentials_path.exists():
            cp = configparser.ConfigParser()
            cp.read(credentials_path)
            for config_key in CREDENTIALS_KEYS:
                self.retrieve_config(cp, config_key)
        if CLICKUP_TOKEN in os.environ:
            self.default_clickup_token = os.environ[CLICKUP_TOKEN]
        if GITHUB_TOKEN in os.environ:
            self.default_github_token = os.environ[GITHUB_TOKEN]
        if not self.default_clickup_token:
            # Check if credentials_path is a temporary file
            if self.is_credentials_path_temporary(credentials_path):
                raise ConfigException(f"Please set the ClickUp token in the {CLICKUP_TOKEN} environment variable")
            else:
                raise ConfigException(
                    f"Please set the ClickUp token in {credentials_path} file or in the {CLICKUP_TOKEN} environment variable"
                )

    def is_credentials_path_temporary(self, credentials_path):
        "Return true if credentials_path is temporary file (i.e. is in tempdir)"
        temp_dir = Path(tempfile.gettempdir())
        return temp_dir in credentials_path.parents

    def load_config(self, base_path):
        "Get project config - load alkemy_workflow.ini file from project root folder"
        git = Git(self)
        self.git_dir = git.get_toplevel(base_path)
        self.config_path = self.git_dir / "alkemy_workflow.ini"
        if self.config_path.exists():
            # Load project config
            cp = configparser.ConfigParser()
            cp.read(self.config_path)
            for config_key in CONFIG_KEYS:
                self.retrieve_config(cp, config_key)

    def retrieve_config(self, cp, config_key):
        section, key = config_key.split("_", 1)
        try:
            setattr(self, config_key, cp[section][key])
        except KeyError:
            pass

    def write_config(self, base_branch=None):
        "Write project config file alkemy_workflow.ini"
        cp = configparser.ConfigParser()
        if self.config_path is None:
            raise GitException("Not a git repository (or any of the parent directories)")
        if self.config_path.exists():
            cp.read(self.config_path)
        if not cp.has_section("git"):
            cp["git"] = {}
        if not cp.has_section("clickup"):
            cp["clickup"] = {}
        if cp["clickup"].get("clickup_status_in_progress") is None:
            cp.set(
                "clickup",
                "status_in_progress",
                self.clickup_status_in_progress,
            )
        if cp["clickup"].get("clickup_status_pr") is None:
            cp.set(
                "clickup",
                "status_pr",
                self.clickup_status_pr,
            )
        if cp["clickup"].get("clickup_status_ma") is None:
            cp.set(
                "clickup",
                "status_ma",
                self.clickup_status_ma,
            )
        if base_branch:
            cp.set("git", "base_branch", base_branch)
        elif not cp.has_option("git", "base_branch"):
            git = Git(self)
            cp.set("git", "base_branch", git.get_current_branch())
        with self.config_path.open("w") as f:
            cp.write(f)

    @classmethod
    def get_credentials_path(cls):
        "Get credential file path"
        try:
            home_dir = Path.home()
            return home_dir / ".alkemy_workflow" / "credentials"
        except Exception:
            with tempfile.NamedTemporaryFile() as f:
                return Path(f.name)

    @classmethod
    def write_credentials(cls, clickup_token, github_token, credentials_path):
        "Write the tokens in the credentials file"
        if clickup_token and not clickup_token.startswith("pk_"):
            raise ConfigException("ClickUp API tokens will always begin with pk_")
        if github_token and not github_token.startswith("gh"):
            raise ConfigException("GitHub tokens will always begin with gh?_ (e.g. ghp_, gho_)")
        credentials_path = credentials_path or cls.get_credentials_path()
        cp = configparser.ConfigParser()
        if credentials_path.exists():
            cp.read(credentials_path)
        if not cp.has_section("default"):
            cp["default"] = {}
        if clickup_token:
            cp.set("default", "clickup_token", clickup_token)
        if github_token:
            cp.set("default", "github_token", github_token)
        credentials_path.parent.mkdir(parents=True, exist_ok=True)
        with credentials_path.open("w") as f:
            cp.write(f)

    @classmethod
    def set_verbose(cls):
        "Turn on vebose output"
        os.environ[AW_VERBOSE] = "true"

    @classmethod
    def is_verbose(cls):
        "True if verbose output is enabled"
        return bool(os.environ.get(AW_VERBOSE))



#!/usr/bin/env python

import sys
import traceback
from datetime import datetime
from pathlib import Path
import click
import pzp
from .exceptions import (
    ClickUpException,
    GenericWarning,
    GenericException,
)
from click.exceptions import MissingParameter
from .config import Config
from .utils import Workflow, VERSION

EXIT_SUCCESS = 0
EXIT_FAILURE = 1
EXIT_PARSER_ERROR = 2

__all__ = ["main"]


def pairwise(iterable):
    return zip(iterable, iterable[1:] + [None])


def prepare_tree(items, enabled=True):
    if not enabled:
        for item in items:
            item["tree"] = str(item["id"])
            yield item
    else:
        level = 0
        for item, next_item in pairwise(items):
            is_last = (next_item is None) or (item["type"] != next_item["type"])
            if level == 0:
                graph = ""
            elif not is_last:
                graph = "├─ "
            else:
                graph = "└─ "
            item["tree"] = f"{' ' * level}{graph}{item.id}"
            if is_last:
                level = level + 1
            yield item


def pick_task(wf, space=None, folder=None, lst=None, task=None):
    "Select a task"
    result = wf.client.query(
        space=space,
        folder=folder,
        lst=lst,
        task=task,
        hierarchy=True,
    )
    fmt = "{tree:20} {label:15} {name:40}"
    header_str = fmt.format(label="Kind/Status", tree="Id", name="Title")
    last_item = None
    while True:
        items = prepare_tree(result, enabled=True)
        item = pzp.pzp(
            items,
            format_fn=lambda item: fmt.format(**item),
            header_str=header_str,
        )
        if item is None:
            return None
        elif item["type"] == "Space":
            item["space"] = {"id": item["id"]}
        elif item["type"] == "Folder":
            item["folder"] = {"id": item["id"]}
        elif item["type"] == "List":
            item["list"] = {"id": item["id"]}
        elif item["type"] == "Task":
            if last_item == item or not item.has_subtasks():
                return item.id
            else:
                item["task"] = {"id": item["id"]}
        elif item["type"] == "Subtask":
            return item.id
        result = wf.client.query(
            space=item.get("space", {}).get("id"),
            folder=item.get("folder", {}).get("id"),
            lst=item.get("list", {}).get("id"),
            task=item.get("task", {}).get("id"),
            hierarchy=True,
        )
        last_item = item


def get_current_task(wf):
    "Get task for current git branch"
    current_branch = wf.git.get_current_branch()
    task_id = current_branch.split("-")[0]
    if task_id == wf.config.git_base_branch:
        raise GenericException(f"Please execute from feature branches, not {wf.config.git_base_branch}")
    return wf.client.get_task_by_id(task_id)


def check_task_status(task, status):
    statuses = task.get_list().get_statuses()
    if status in statuses:
        return True
    else:
        click.secho(
            f"Warning: Status '{status}' not found. Valid statuses are: {', '.join(statuses)}",
            fg="yellow",
        )
        return False


@click.group("cli")
@click.version_option(VERSION)
@click.pass_context
@click.option(
    "-C",
    "--cwd",
    help="Run as if git was started in <path> instead of the current working directory",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
)
@click.option(
    "--credentials-path",
    help="Credentials file path",
    type=click.Path(exists=False, file_okay=True, dir_okay=False),
)
@click.option(
    "-v", "--verbose",
    help="Verbose output",
    default=False,
    is_flag=True,
)
def cli(ctx, cwd, credentials_path, verbose):
    if verbose:
        Config.set_verbose()
    ctx.obj = Workflow(cwd, credentials_path)


@cli.command("spaces")
@click.option("--filter", help="Filter spaces by name")
@click.option("--headers/--noheaders", default=True, help="Show/hide headers")
@click.pass_context
def cmd_spaces(ctx, filter, headers):
    """
    List spaces

    Example: aw spaces --filter 'test*' --noheaders
    """
    wf = ctx.obj
    result = wf.client.query(filter_type="Space", filter_name=filter)
    fmt = "{id:15} {name:40}"
    if headers:
        print(fmt.format(id="Id", name="Space"))
        print("-" * 70)
    for item in result:
        print(fmt.format(**item))


@cli.command("folders")
@click.option("--space", help="Space name", required=True)
@click.option("--filter", help="Filter folders by name")
@click.option("--headers/--noheaders", default=True, help="Show/hide headers")
@click.pass_context
def cmd_folders(ctx, space, filter, headers):
    """
    List folders from a space

    Example: aw folders --space 'test' --filter 'abc*' --noheaders
    """
    wf = ctx.obj
    result = wf.client.query(space=space, filter_type="Folder", filter_name=filter)
    fmt = "{id:15} {name:40}"
    if headers:
        print(fmt.format(id="Id", name="Folders"))
        print("-" * 70)
    for item in result:
        print(fmt.format(**item))


@cli.command("lists")
@click.option("--space", help="Space name")
@click.option("--folder", help="Folder name")
@click.option("--filter", help="Filter lists by name")
@click.option("--headers/--noheaders", default=True, help="Show/hide headers")
@click.pass_context
def cmd_lists(ctx, space, folder, filter, headers):
    """
    List lists from a space or folder

    Example: aw lists --space 'test' --filter 'abc*' --noheaders
    """
    wf = ctx.obj
    if not space and not folder:
        raise click.ClickException("Missing option '--space' or '--folder'")
    result = wf.client.query(
        space=space,
        folder=folder,
        filter_type="List",
        filter_name=filter,
    )
    fmt = "{id:15} {name:40}"
    if headers:
        print(fmt.format(id="Id", name="Lists"))
        print("-" * 70)
    for item in result:
        print(fmt.format(**item))


@cli.command("tasks")
@click.option("--space", help="Space name")
@click.option("--folder", help="Folder name")
@click.option("--list", help="List name")
@click.option("--task", help="Task id")
@click.option("--filter", help="Filter tasks by name")
@click.option("--headers/--noheaders", default=True, help="Show/hide headers")
@click.pass_context
def cmd_tasks(ctx, space, folder, list, task, filter, headers):
    """
    List tasks from a list or subtask
    """
    wf = ctx.obj
    if not list and not task:
        raise click.ClickException("Missing option '--list' or '--task'")
    result = wf.client.query(space=space, folder=folder, lst=list, task=task, filter_name=filter)
    fmt = "{label:15} {id:15} {name:40}"
    if headers:
        print(fmt.format(id="Id", label="Status", name="Title"))
        print("-" * 70)
    for item in result:
        print(fmt.format(**item))


@cli.command("ls")
@click.option("--space", help="Space name")
@click.option("--folder", help="Folder name")
@click.option("--list", help="List name")
@click.option("--task", help="Task id")
@click.option("--filter", help="Filter tasks by name")
@click.option("--headers/--noheaders", default=True, help="Show/hide headers")
@click.option("--hierarchy/--nohierarchy", default=True, help="Show/hide hierarchy")
@click.pass_context
def cmd_ls(ctx, space, folder, list, task, filter, headers, hierarchy):
    """
    List spaces --> folders --> lists --> tasks --> subtasks

    Example: aw ls --space 'test' --noheaders
    """
    wf = ctx.obj
    result = wf.client.query(
        space=space,
        folder=folder,
        lst=list,
        task=task,
        filter_name=filter,
        hierarchy=hierarchy,
    )
    fmt = "{tree:20} {label:15} {name:40}"
    if headers:
        print(fmt.format(label="Kind/Status", tree="Id", name="Title"))
        print("-" * 70)
    for item in prepare_tree(result, enabled=hierarchy):
        print(fmt.format(**item))


@cli.command("branch")
@click.argument("task_id", required=False)
@click.option("--repo", help="Remote repository URL")
@click.pass_context
def cmd_branch(ctx, task_id, repo):
    """
    Open a task and create a new git branch

    Example: aw branch '#12abcd45'
    """
    wf = ctx.obj
    if task_id is None:
        # Show task picker
        task_id = pick_task(wf=wf)
        if task_id is None:
            raise MissingParameter(ctx=ctx, param_hint="'TASK_ID'", param_type="argument")
    task = wf.client.get_task_by_id(task_id)
    if repo:
        # Create a new remote branch
        branch_already_exists = wf.github.create_branch(repo, task.branch_name)
    else:
        # Create a new local branch a switch to it
        branch_already_exists = wf.git.create_branch(task.branch_name)
    # Update the task
    task_update = {}  # task fields to be updated
    # Start date
    if not task.get("start_date"):
        task_update["start_date"] = int(datetime.utcnow().timestamp() * 1000)
    # Task assignee
    current_user = wf.client.get_user()
    if current_user["id"] not in [x["id"] for x in task["assignees"]]:
        task_update["assignees"] = {"add": [current_user["id"]]}
    # Status
    new_status = wf.config.clickup_status_in_progress
    if task["status"].get("type") == "open" and check_task_status(task, new_status):
        task_update["status"] = new_status
    # Update the task
    task.update_task(**task_update)
    # Post a comment
    if not branch_already_exists:
        github_url = wf.git.get_github_url(task.branch_name, repo)
        if github_url:
            comment = f"Branch [{task.branch_name}]({github_url})"
        else:
            comment = f"Branch {task.branch_name}"
        task.post_task_comment(comment)
    click.secho(f"Branch {task.branch_name}", fg="green")


@cli.command("commit")
@click.option("-m", "--message", help="Commit message")
@click.pass_context
def cmd_commit(ctx, message):
    """
    Create a new commit an the current feature branch

    Example: aw commit
    """
    wf = ctx.obj
    task = get_current_task(wf)
    # Prepare the commit message
    if message:
        commit_message = f"[{task['id']}] {task['name']} - {message}"
    else:
        commit_message = f"[{task['id']}] {task['name']}"
    # Commit
    print(wf.git.run("commit", "-m", commit_message))


@cli.command("pr")
@click.option("--repo", help="Remote repository URL")
@click.argument("task_id", required=False)
@click.pass_context
def cmd_pr(ctx, repo, task_id):
    """
    Push local commits to the remote branch and create a pull request on GitHub

    Example: aw pr
    """
    wf = ctx.obj
    if task_id:
        # Task id argument
        task = wf.client.get_task_by_id(task_id)
    else:
        # Get task from current git branch
        task = get_current_task(wf)
    # Push
    if not repo:
        wf.git.push()
    # Create the pull request
    repo = repo or wf.git.get_remote_url()
    title = f"[{task['id']}] {task['name']}"
    response = wf.github.create_pull_request(repo, task.branch_name, title)
    click.secho(f"Pull request created\n{response['html_url']}", fg="green")
    # Update task status
    if wf.config.clickup_status_pr:
        new_status = wf.config.clickup_status_pr
        if check_task_status(task, new_status):
            task.update_task(status=new_status)


@cli.command("lr")
@click.option("--repo", help="Remote repository URL")
@click.option("--headers/--noheaders", default=True, help="Show/hide headers")
@click.argument("task_id", required=False)
@click.pass_context
def cmd_lr(ctx, repo, headers, task_id):
    """
    List pull requests on repository

    Example: aw lr
    """
    wf = ctx.obj
    # List the pull request
    repo = repo or wf.git.get_remote_url()
    result = wf.github.list_pull_request(repo)

    fmt = "{number:6} {title:50} {diff_url}"
    if headers:
        print(fmt.format(number="Pr.num", title="Title", diff_url="Diff url"))
        print("-" * 70)
    for item in result:
        print(fmt.format(**item))


@cli.command("merge")
@click.option("--repo", help="Remote repository URL")
@click.option("--pr_nr", help="Pull request number")
@click.argument("task_id", required=False)
@click.pass_context
def cmd_ma(ctx, repo, pr_nr, task_id):
    """
    Merge a certain pull request

    Example: aw merge --pr_nr 1
    """
    wf = ctx.obj
    # Merge the pull request
    repo = repo or wf.git.get_remote_url()
    if pr_nr is None:
        # Show pull request picker
        fmt = "{number:6} {title:50} {diff_url}"
        pull_requets = wf.github.list_pull_request(repo)
        pr = pzp.pzp(
            pull_requets,
            format_fn=lambda item: fmt.format(**item),
            fullscreen=False,
        )
        if pr is None:
            raise MissingParameter(ctx=ctx, param_hint="'--pr_nr'", param_type="option")
        else:
            pr_nr = pr['number']
    wf.github.merge_pull_request(repo, pr_nr)
    if task_id:
        # Task id argument
        task = wf.client.get_task_by_id(task_id)
    else:
        # Get task from current git branch
        task = get_current_task(wf)
    # Update status
    if wf.config.clickup_status_ma:
        new_status = wf.config.clickup_status_ma
        if check_task_status(task, new_status):
            task.update_task(status=new_status)


@cli.command("get-status")
@click.argument("task_id", required=False)
@click.pass_context
def cmd_get_status(ctx, task_id):
    """
    Get task status

    Example: aw get-status '#12abcd45'
    """
    wf = ctx.obj
    if task_id is None:
        # Show task picker
        task_id = pick_task(wf=wf)
        if task_id is None:
            raise MissingParameter(ctx=ctx, param_hint="'TASK_ID'", param_type="argument")
    task = wf.client.get_task_by_id(task_id)
    space = task.get_space()
    style = lambda x: click.style(x, "cyan")
    print(
        f"""\
{style('Task id:')}  {task.task_id}
{style('Status:')}   {task.status}
{style('Space:')}    {space['name']}
{style('Folder:')}   {task['folder']['name']}
{style('List:')}     {task['list']['name']}
{style('Title:')}    {task['name']}"""
    )


@cli.command("set-status")
@click.argument("task_id", required=False)
@click.argument("status", required=False)
@click.pass_context
def cmd_set_status(ctx, task_id, status):
    """
    Change task status

    Example: aw set-status '#12abcd45' 'done'
    """
    wf = ctx.obj
    if task_id is None:
        # Show task picker
        task_id = pick_task(wf=wf)
        if task_id is None:
            raise MissingParameter(ctx=ctx, param_hint="'TASK_ID'", param_type="argument")
    task = wf.client.get_task_by_id(task_id)
    if status is None:
        # Show status picker
        statuses = task.get_list().get_statuses()
        status = pzp.pzp(
            statuses,
            fullscreen=False,
        )
        if status is None:
            raise MissingParameter(ctx=ctx, param_hint="'STATUS'", param_type="argument")
    # Update the task
    try:
        task.update_task(status=status)
        click.secho(f"Task {task_id} status changed to {status}", fg="green")
    except ClickUpException:
        statuses = task.get_list().get_statuses()
        raise GenericException(f"Error setting status. Valid statuses are: {', '.join(statuses)}")


@cli.command("configure")
@click.option("--clickup-token", help="ClickUp API token")
@click.option("--github-token", help="GitHub token")
@click.pass_context
def cmd_configure(ctx, clickup_token, github_token):
    """
    Configures credentials (Clickup and GitHub tokens).

    Example: aw configure
    """
    wf = ctx.obj
    prompt = "ClickUP API token: "
    clickup_token = clickup_token or input(prompt).strip()
    prompt = "GitHub token: "
    github_token = github_token or input(prompt).strip()
    Config.write_credentials(clickup_token, github_token, wf.credentials_path)
    wf.client.get_user()
    click.secho("ClickUP API token verified", fg="green")
    if wf.config.default_github_token:
        wf.github.get_user()
        click.secho("GitHub token verified", fg="green")


@cli.command("init")
@click.option("--base-branch", help="Git base branch")
@click.pass_context
def cmd_init(ctx, base_branch):
    """
    Write alkemy_workflow.ini config file into the project root folder

    Example: aw init --base-branch=main
    """
    wf = ctx.obj
    wf.config.write_config(base_branch=base_branch)


def main(argv=None):
    if argv:
        prog_name = Path(argv[0]).name
        args = argv[1:]
    else:
        args = None
        prog_name = "aw"
    try:
        cli(prog_name=prog_name, args=args)
        return EXIT_SUCCESS
    except SystemExit as err:
        return err.code
    except GenericWarning as ex:
        click.secho(f"{prog_name}: {ex}", fg="yellow")
        if Config.is_verbose():
            traceback.print_exc(file=sys.stdout)
        return EXIT_SUCCESS
    except GenericException as ex:
        click.secho(f"{prog_name}: {ex}", fg="red")
        if Config.is_verbose():
            traceback.print_exc(file=sys.stdout)
        return EXIT_FAILURE


if __name__ == "__main__":
    sys.exit(main(sys.argv))
